import tkinter
from version import *

from vault import generate_password
from vault import VALID_PASSWORD_TYPES

class Dialog(tkinter.Toplevel):
    """
    Parent Dialog frame, inherit and override
    body and apply methods.
    """
    def __init__(self, parent, title=None, initial_data=None):
        super().__init__(parent)
        self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent

        self.result = None

        body = tkinter.ttk.Frame(self)
        self.initial_focus = self.body(body, initial_data)
        body.pack(padx=5, pady=5, fill=tkinter.BOTH)

        self.buttonbox()

        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    def body(self, master, initial_data=None):
        """Body of dialog, override in child classes."""
        pass

    def buttonbox(self):
        """Standard OK and Cancel buttons."""
        box = tkinter.ttk.Frame(self)

        w = tkinter.ttk.Button(
            box, text="OK", width=10, command=self.ok, default=tkinter.ACTIVE)
        w.pack(side='left', padx=5, pady=5)
        w = tkinter.ttk.Button(
            box, text="Cancel", width=10, command=self.cancel)
        w.pack(side='left', padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()
        return box

    def ok(self, event=None):
        """On OK button press."""
        if not self.validate():
            self.initial_focus.focus_set()
            return

        self.withdraw()
        self.update_idletasks()

        self.apply()

        self.cancel()

    def cancel(self, event=None):
        """On Cancel button press."""
        self.parent.focus_set()
        self.destroy()

    def validate(self):
        """
        Validate action on close, defaults to True if not overridden.
        """
        return 1

    def apply(self):
        """Apply result on OK button press, override in child class."""
        pass

class AddPassword(Dialog):
    def body(self, master, initial_data):
        """Body of set key dialog."""
        for index, key in enumerate(('system',
                                     'username',
                                     'password',
                                     'notes')):
            setattr(self, key, tkinter.StringVar())
            e = LabelEntry(master,
                           width=50,
                           label=key.title(),
                           textvariable=getattr(self, key))
            e.pack()
            if initial_data:
                getattr(self, key).set(initial_data[index])

        f = tkinter.Frame(master)
        f.pack()
        pw_type = tkinter.StringVar()
        pw_type.set(VALID_PASSWORD_TYPES[0])
        pw_len = tkinter.IntVar()
        pw_len.set(10)
        gen_pass = tkinter.OptionMenu(f, pw_type, *VALID_PASSWORD_TYPES)
        gen_pass.configure(width=20)
        length = tkinter.OptionMenu(f, pw_len, *range(4,16))
        length.configure(width=2)
        
        gen_pass.pack(side='left')
        length.pack(side='left')
        b = tkinter.Button(
            f,
            text='Generate Password',
            command=lambda:self.password.set(
                generate_password(
                    pw_type.get(), pw_len.get())))
        b.pack(side='left')
                
        
    def apply(self):
        """Set result upon OK button press."""
        self.result = (self.system.get(),
                       self.username.get(),
                       self.password.get(),
                       self.notes.get())
        

class SetupSSH(Dialog):
    def body(self, master, initial_data):
        """Body of set key dialog."""
        for index, key in enumerate(('host',
                                     'port',
                                     'username',
                                     'password')):
            setattr(self, key, tkinter.StringVar())
            e = LabelEntry(master,
                           width=50,
                           label=key.title().replace('_', ' '),
                           textvariable=getattr(self, key))
            e.pack()
            if key == 'password':
                e.set_config(show='*')
            if initial_data:
                try:
                    getattr(self, key).set(initial_data[index])
                except IndexError:
                    pass
        
    def apply(self):
        """Set result upon OK button press."""
        self.result = [getattr(self, key).get() for
                       key in ('host', 'port', 'username', 'password')]

class SetupFiles(Dialog):
    def body(self, master, initial_data):
        # Checkboxes.
        for key in ('use_steganography', 'clear_on_exit'):
            setattr(self, key, tkinter.BooleanVar())
            c = tkinter.Checkbutton(
                master,
                text=key.replace('_', ' ').title(),
                variable=getattr(self, key),
                anchor='w')
            c.pack(expand=1, fill=tkinter.X)
        if initial_data:
            self.use_steganography.set(initial_data.get(
                'use_steganography', False))
            self.clear_on_exit.set(initial_data.get('clear_on_exit', True))

        # Entries
        for key, name in (('file_location', 'Password file location'),
                          ('original_file', 'Steganography original file location')):
            setattr(self, key, tkinter.StringVar())
            if initial_data:
                getattr(self, key).set(initial_data.get(key, ''))

            e = LabelEntry(master,
                           width=50,
                           label=name,
                           textvariable=getattr(self, key))
            e.pack()
            

    def apply(self):
        self.result = {key: getattr(self, key).get() for
                       key in ('file_location', 'original_file',
                               'use_steganography', 'clear_on_exit')}

class About(Dialog):
    def body(self, master, _):
        for name, var in (('', '__summary__'),
                          ('Version: ', '__version__'),
                          ('Author: ', '__author__'),
                          ('Contact: ', '__email__'),
                          ('Licence: ', '__license__'),
                          ('', '__uri__')):
            l = tkinter.Label(master, text=f'{name} {eval(var)}')
            l.pack(fill=tkinter.X, expand=1)

    def buttonbox(self):
        """Standard OK and Cancel buttons."""
        box = tkinter.ttk.Frame(self)

        w = tkinter.ttk.Button(
            box, text="OK", width=10, command=self.ok, default=tkinter.ACTIVE)
        w.pack(side='left', padx=5, pady=5)
        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()
        return box

class Box(tkinter.Frame):
    """Frame box to make widgets align."""
    def __init__(self, master, widget_type, *args, **kwargs):
        self.master = master
        side = kwargs.pop('side', 'bottom')
        super().__init__(master)
        self.widget = getattr(tkinter, widget_type)(self, *args, **kwargs)
        self.widget.pack(side=side)

    def config(self, *args, **kwargs):
        self.widget.config(*args, **kwargs)

class StatusBar(tkinter.Label):
    def __init__(self, master, *args, **kwargs):
        self.default_color = master.cget('bg')
        self.master = master
        self._data = kwargs.pop('textvariable', tkinter.StringVar())
        super().__init__(master, *args, textvariable=self._data, **kwargs)

    def set(self, data, color=None):
        color = color or self.default_color
        self._data.set(data)
        self.config(bg=color)
        self.master.update_idletasks()

class LabelEntry(tkinter.Frame):
    def __init__(self, master, *args, **kwargs):
        self.master = master
        label = kwargs.pop('label', '')
        super().__init__(master)
        self._label = tkinter.Label(self, text=label)
        self._entry = tkinter.Entry(self, *args, **kwargs)
        self._label.pack(side='top', fill=tkinter.X)
        self._entry.pack(side='top', fill=tkinter.X)

    @property
    def label(self, value):
        self._label.set(value)

    def set_config(self, **kwargs):
        self._entry.configure(**kwargs)

    def focus_set(self, *args, **kwargs):
        self._entry.focus_set(*args, **kwargs)

    def bind(self, *args, **kwargs):
        self._entry.bind(*args, **kwargs)



import pickle
import pytest

@pytest.fixture
def vault_data_locked():
    return {'salt': b'\xa0\x9b\xb4\xbe\xd1\xb7\xe0t\x11\x9a\xd2\xe2\x10\x04\x03\x7f',
            'iterations': 1000000,
            'vault': b'gAAAAABfsq0JevttlDtLkaD2QfQn96QDOFOyh1egDiatqTlDnE98zHJx9SbKLRAPuZLPl4X3JOZZq2FOA5sfawk6TSRTnIdQv-Ik3MCr_ZFGxVZ85EW6P9c='}

@pytest.fixture
def vault_data_unlocked():
    return {'salt': b'\xa0\x9b\xb4\xbe\xd1\xb7\xe0t\x11\x9a\xd2\xe2\x10\x04\x03\x7f',
            'iterations': 1000000,
            'vault': ['test']}

@pytest.fixture
def locked_file(tmpdir, vault_data_locked):
    file = tmpdir.join('testfile.txt')
    with open(file, 'bw') as fh:
        pickle.dump(vault_data_locked, fh)
    return file



import ast
import base64
import os
import pickle
import random
import string

from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

VALID_PASSWORD_TYPES = ('alpha',
                        'alphanum',
                        'alphanumspecial',
                        'mobilealpha',
                        'mobilealphanum',
                        'mobilealphanumspecial',
                        'numerical')

class Vault():
    def __init__(self, data_file=None):
        self._locked = False
        if data_file:
            self.load_file(data_file)
        else:
            self.data = {'salt': os.urandom(16),
                         'iterations': 1000000,
                         'vault': []}

    @property
    def locked(self):
        return self._locked
    
    def load_data(self, data):
        self.data = pickle.loads(data)
        self._locked = True

    def save_data(self):
        return pickle.dumps(self.data)
    
    def load_file(self, fh):
        self.data = pickle.load(fh)
        self._locked = True

    def save_file(self, fh):
        pickle.dump(self.data, fh)

    def load_clear(self, fh):
        if self.locked:
            raise VaultError('Vault is locked, unlock first!')
        self._locked = False
        for row in fh:
            self.add(ast.literal_eval(row.strip()))

    def save_clear(self, fh):
        if self.locked:
            raise VaultError('Vault is locked, please unlock first')
        for obj in self.data['vault']:
            print(repr(obj), file=fh)
        
    def lock(self, password):
        if self._locked: return
        data = pickle.dumps(self.data['vault'])
        key = self.create_key(password,
                              self.data['salt'],
                              self.data.get('iterations', 1000000))
        self.data['vault'] = Fernet(key).encrypt(data)
        self._locked = True

    def unlock(self, password):
        if not self._locked: return
        key = self.create_key(password,
                              self.data['salt'],
                              self.data.get('iterations', 1000000))
        try:
            self.data['vault'] = pickle.loads(Fernet(key).decrypt(
                self.data['vault']))
            self._locked = False
        except InvalidToken:
            pass
        
    def create_key(self, password, salt, iterations=1000000):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=iterations,
            backend=default_backend()
            )
        return base64.urlsafe_b64encode(
            kdf.derive(bytearray(password, 'utf-8')))

    def get_objects(self):
        return self.data['vault']

    def set_objects(self, objs):
        self.data['vault'] = objs
    
    def add(self, obj):
        self.data['vault'].append(obj)

    def remove_password(self, obj):
        self.data['vault'].remove(obj)
        

class VaultError(Exception):
    pass
        
def generate_password(password_type='alpha', n=10):
    alphabets = {'alpha': (string.ascii_letters, 'isupper', 'islower'),
                 'alphanum': (string.ascii_letters + string.digits, 'isupper', 'islower','isdigit'),
                 'alphanumspecial': (string.ascii_letters + string.digits + '!"#¤%&/()=?', 'isupper', 'islower','isdigit'),
                 'mobilealpha': (string.ascii_lowercase, 'islower'),
                 'mobilealphanum': (string.ascii_lowercase, 'islower'),
                 'mobilealphanumspecial': (string.ascii_lowercase, 'islower'),
                 'numerical': (string.digits, 'isdigit')}
    if not password_type in alphabets:
        raise KeyError(f'Password type has to be one of the following: {", ".join(alphabets.keys())} it is "{password_type}".')
    while True:
        alphabet = alphabets[password_type][0]
        password = random.choices(alphabet, k=n)
        if password_type.startswith('mobile'):
            password[0] = password[0].upper()
        if 'mobilealphanum' in password_type :
            password[-2] = random.choice(string.digits)
        if password_type == 'mobilealphanumspecial':
            password[-1] = random.choice('!"#¤%&/()=?')
            
        for test in alphabets[password_type][1:]:
            if not [c for c in password if getattr(c, test)()]:
                break
        else:
            if password_type == 'alphanumspecial':
                if not [c for c in password if not c.isalpha()]:
                    continue
            break
    return ''.join(password)



import io
from PIL import Image
import pickle

def write(fh, original, data):
    with Image.open(original) as image:
        mask = [int(y) for y in ''.join([str(format(x, 'b')).zfill(8) for x in data])]
        new_band = []
        for index, x in enumerate(image.getdata(0)):
            new_band.append((x + (mask.pop(0) if mask else 2)%255))

        r,g,b = image.split()
        r.putdata(new_band)
        i = Image.merge(image.mode, (r, g, b))
        i.save(fh, 'png')

def read(fh, original):
    with Image.open(original) as orig, Image.open(fh) as mask:
        orig = list(orig.getdata(0))
        mask = list(mask.getdata(0))
        result = []
        for x, y in zip(mask, orig):
            value = x - y 
            if value in (2, -254):
                break
            if value < 0:
                value = 1
            result.append(str(value))
        return int(''.join(result), 2).to_bytes(len(result) // 8, byteorder='big')



import os
import paramiko
import tkinter.messagebox

class RemoteFile():
    """
    Class to handle remote files through ssh.
    """
    def __init__(self, host, port, username, password, filepath, data_dir, *args, **kwargs):
        self.ssh = paramiko.SSHClient()
        try:
            self.ssh.load_host_keys(os.path.join(data_dir, '.know_hosts'))
        except FileNotFoundError:
            os.makedirs(data_dir, exist_ok=True)
            with open(os.path.join(data_dir, '.know_hosts'), 'w'):
                pass
        self._connect(host, port, username, password)
        self.filepath = filepath
        self.sftp = None
        self.stat = None
        self.fh = self._open(*args, **kwargs)
             
    def __enter__(self):
        return self.fh

    def __exit__(self, *args):
        self.close()

    def _connect(self, host, port, username, password, unknown_host=False):
        try:
            if unknown_host:
                self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy)
            self.ssh.connect(host, int(port), username, password)
        except paramiko.ssh_exception.SSHException as err:
            if str(err) == 'Authentication failed.':
                tkinter.messagebox.showerror('Authentication failed.',
                                             'Authentication failed.')
                self.ssh.close()
                return
            if tkinter.messagebox.askokcancel(
                "Unknown host", "Unknown host!\n"
                "Do you want to continue connecting?"):
                self._connect(host, port, username, password, True)
            else:
                self.ssh.close()

    def _makedirs(self, path):
        current_path = ''
        for p in path.split('/'):
            current_path = (current_path and f'{current_path}/{p}') or p
            try:
                self.sftp.stat(current_path)
            except FileNotFoundError:
                self.sftp.mkdir(current_path)

    def close(self):
        if self.stat:
            self.sftp.utime(self.filepath, (self.stat.st_atime,
                                            self.stat.st_mtime))
        for con in ('fh', 'sftp', 'ssh'):
            try:
                getattr(self, con).close()
            except:
                pass
                
    def _open(self, *args, **kwargs):
        """Open remote path."""
        path = kwargs.pop('path', self.filepath)
        transport = self.ssh.get_transport()
        try:
            transport.send_ignore()
        except:
            return            
        self.sftp = self.ssh.open_sftp()
        self._makedirs(os.path.dirname(path))
        try:
            self.stat = self.sftp.stat(path)
            return self.sftp.open(path, *args, **kwargs)
        except FileNotFoundError:
            tkinter.messagebox.showerror('File not found!',
                                         f'Unable to find the file {self.filepath}')



import pickle

from tests.fixtures import *
import acid_vault.vault as vault

def test_unlock(vault_data_locked):
    v = vault.Vault()
    v.data = vault_data_locked
    v._locked = True
    v.unlock('testpass')
    assert 'test' in v.data['vault']

def test_already_unlocked(vault_data_unlocked):
    v = vault.Vault()
    v.data = vault_data_unlocked
    v.unlock('testpass')
    assert 'test' in v.data['vault']
    
def test_lock(vault_data_unlocked):
    v = vault.Vault()
    v.data = vault_data_unlocked
    v.lock('testpass')
    try:
        assert 'test' not in v.data['vault']
    except TypeError:
        # v.data is a binary string so should trhow error
        pass
    assert b'test' not in v.data['vault']
    
def test_load_data(vault_data_locked):
    data = pickle.dumps(vault_data_locked)
    v = vault.Vault()
    v.load_data(data)
    v.unlock('testpass')
    assert 'test' in v.data['vault']

def test_save_data(vault_data_locked):
    v = vault.Vault()
    v.data = vault_data_locked
    data = v.save_data()
    v2 = vault.Vault()
    v2.load_data(data)
    v2.unlock('testpass')
    assert 'test' in v2.data['vault']

def test_load_file(locked_file):
    v = vault.Vault()
    with open(locked_file, 'rb') as fh:
        v.load_file(fh)
    v.unlock('testpass')
    assert 'test' in v.data['vault']

def test_save_file(vault_data_locked, tmpdir):
    v = vault.Vault()
    v.data = vault_data_locked
    f = tmpdir.join('testfile.txt')
    with open(f, 'wb') as fh:
        v.save_file(fh)
    v2 = vault.Vault()
    with open(f, 'rb') as fh:
        v2.load_file(fh)
    v2.unlock('testpass')
    assert 'test' in v2.data['vault']

def test_load_clear(tmpdir, vault_data_unlocked):
    v = vault.Vault()
    v._locked = False
    v.data = vault_data_unlocked
    v.data['vault'] = []
    f = tmpdir.join('testfile.txt')
    with open(f, 'w') as fh:
        print('("test", )', file=fh)
    with open(f, 'r') as fh:
        v.load_clear(fh)
    assert ("test", ) in v.data['vault']
    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

meta_data = {}
with open('acid_vault/version.py') as fh:
    exec(fh.read(), meta_data)

setuptools.setup(
    name=meta_data['__title__'],
    version=meta_data['__version__'],
    author=meta_data['__author__'],
    author_email=meta_data['__email__'],
    description=meta_data['__summary__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=meta_data["__uri__"],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['cryptography>=3.0',
                      'paramiko>=2.7',
                      'pillow>=7.0']
)



import os
import platform

def data_dir():
    system = platform.system()
    if system == 'Windows':
        path = os.path.join('appdata', 'local', 'vault')
    elif system == 'Linux':
        path = '.vault'
    else:
        path = ''
    path = os.path.expanduser(os.path.join('~', path))
    os.makedirs(path, exist_ok=True)
    return path
